type 'a t = 'a list
(* hd : 'a list -> 'a *)
let hd = function x::_ -> x
                | _    -> failwith "empty list"
(* tl : 'a list -> 'a list *)
let tl = function _::xs -> xs
                | _     -> failwith "empty list"
(* append : 'a list -> 'a list -> 'a list *)
let rec append xs ys = match xs with []    -> ys
                                   | x::xs -> x::(append xs ys)
(* length : 'a list -> int *)
let rec length = function []    -> 0
                        | _::xs -> length xs + 1
(* init : 'a -> int -> 'a list *)
let rec init i l = if l < 1 then [] else i::init i (l - 1)
(* take : int -> 'a list -> 'a list *)
let rec take n xs = if n < 1 then [] else match xs with
 | []    -> []
 | x::xs -> x::take (n - 1) xs
(* drop : int -> 'a list -> 'a list *)
let rec drop n xs = if n < 1 then xs else match xs with
 | []    -> []
 | _::xs -> drop (n - 1) xs
(* split_at : int -> 'a list -> ('a list * 'a list *)
let split_at n xs = (take n xs, drop n xs)
(* map : ('a -> 'b) -> 'a list -> 'b list *)
let rec map f = function []    -> []
                       | x::xs -> let r = f x in r::map f xs
(* foldr : ('a -> 'b -> 'b) -> 'a t -> 'b -> 'b *)
let rec foldr f xs b = match xs with
 | []    -> b
 | x::xs -> f x (foldr f xs b)
(* foldl : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a *)
let rec foldl f b = function []    -> b
                           | x::xs -> foldl f (f b x) xs
(* filter : ('a -> bool) -> 'a list -> 'a list *)
let rec filter p = function
 | []    -> []
 | x::xs -> if p x then x::filter p xs else filter p xs
(* zip2_with : ('a -> 'b -> c) -> 'a list -> 'b list -> 'c list *)
let rec zip2_with f xs ys = match (xs,ys) with
 | (x::xs,y::ys) -> f x y :: zip2_with f xs ys
 | _             -> []
(* zip2 : 'a list -> 'b list -> ('a * 'b)list *)
let zip2 xs ys = zip2_with (fun x y -> (x,y)) xs ys
(* zip4_with : ('a -> 'b -> 'c -> 'd -> 'e) -> 'a t -> 'b t -> 'c t -> 'd t -> 'e t *)
let rec zip4_with f ws xs ys zs = match (ws,xs,ys,zs) with
 | (w::ws,x::xs,y::ys,z::zs) -> f w x y z :: zip4_with f ws xs ys zs
 | _                         -> []
(* zip4 : 'a t -> 'b t -> 'c t -> 'e t -> ('a * 'b * 'c * 'd)t *)
let zip4 ws xs ys zs = zip4_with (fun w x y z -> (w,x,y,z)) ws xs ys zs
(* foldr1 : ('a -> 'a -> 'a) -> 'a list -> 'a *)
let rec foldr1 f = function []    -> failwith "empty list"
                          | [x]   -> x
			  | x::xs -> f x (foldr1 f xs)
(* join : 'a list -> 'a list list -> 'a list *)
let join delim = foldr1 (fun xs ys -> xs@delim@ys)
(* until : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a *)
let rec until p f x = if p x then x else until p f (f x)
(* take_while : ('a -> bool) -> 'a list -> 'a list *)
let rec take_while p = function
 | []    -> []
 | x::xs -> if p x then x :: take_while p xs else []
(* drop_while : ('a -> bool) -> 'a list -> 'a list *)
let rec drop_while p = function
 | []            -> []
 | x::xs as list -> if p x then drop_while p xs else list
(* span : ('a -> bool) -> 'a list -> ('a list,'a list) *)
let rec span p = function
 | [] -> ([],[])
 | x::xs as list ->
  if p x then let (ys,zs) = span p xs in (x::ys,zs)
         else ([],list)
(* rev : 'a list -> 'a list *)
let rev xs =
 let rec rev acc = function []    -> acc
                          | x::xs -> rev (x::acc) xs
 in
 rev [] xs
(* nth : int -> 'a list -> 'a *)
let rec nth i xs =
 if i < 0 then failwith "negative index" else match (xs,i) with
  | ([],_)    -> failwith "index too large"
  | (x::_,0)  -> x
  | (_::xs,i) -> nth (i-1) xs
(* last : 'a list -> 'a *)
let rec last = function
 | []    -> failwith "empty list"
 | [x]   -> x
 | _::xs -> last xs
(* concat : 'a list list -> 'a list *)
let concat xs = foldr (@) xs []
(*  *)
